use std::num::ParseIntError;

use axum::{
    http::{StatusCode, header::ToStrError},
    response::{IntoResponse, Response},
    Json,
};
use eyre::Report;
use serde::Serialize;
use serde_json::json;
use tracing::error;

use crate::FoxError;

pub struct ApiError {
    status: StatusCode,
    code: ErrorCode,
    message: String,
}

impl IntoResponse for ApiError {
    fn into_response(self) -> Response {
        (
            self.status,
            Json(json!({
                "status": self.status.as_u16(),
                "code": self.code,
                "message": self.message,
            })),
        )
            .into_response()
    }
}

#[derive(Serialize)]
#[serde(rename_all = "SCREAMING_SNAKE_CASE")]
pub enum ErrorCode {
    InternalServerError,
    ObjectNotFound,
    InvalidServer,
    InvalidHeader,
    InvalidDate,
    InvalidSignature,
    MissingSignature,
}

impl From<sqlx::Error> for ApiError {
    fn from(err: sqlx::Error) -> Self {
        ApiError {
            status: StatusCode::INTERNAL_SERVER_ERROR,
            code: ErrorCode::InternalServerError,
            message: if cfg!(debug_assertions) {
                format!("Database error: {}", err)
            } else {
                "Database error".into()
            },
        }
    }
}

impl From<Report> for ApiError {
    fn from(err: Report) -> Self {
        match err.downcast_ref::<FoxError>() {
            Some(e) => return (*e).into(),
            None => {}
        };

        error!("Error in handler: {}", err);
        ApiError {
            status: StatusCode::INTERNAL_SERVER_ERROR,
            code: ErrorCode::InternalServerError,
            message: if cfg!(debug_assertions) {
                format!("Internal server error: {}", err)
            } else {
                "Internal server error".into()
            },
        }
    }
}

impl From<FoxError> for ApiError {
    fn from(err: FoxError) -> Self {
        match err {
            FoxError::NotFound => ApiError {
                status: StatusCode::NOT_FOUND,
                code: ErrorCode::ObjectNotFound,
                message: "Object not found".into(),
            },
            FoxError::InvalidServer => ApiError {
                status: StatusCode::BAD_REQUEST,
                code: ErrorCode::InvalidServer,
                message: "Invalid domain or server".into(),
            },
            FoxError::MissingSignature => ApiError {
                status: StatusCode::BAD_REQUEST,
                code: ErrorCode::MissingSignature,
                message: "Missing signature".into(),
            },
            FoxError::InvalidSignature => ApiError {
                status: StatusCode::BAD_REQUEST,
                code: ErrorCode::InvalidSignature,
                message: "Invalid signature".into(),
            },
            _ => ApiError {
                status: StatusCode::INTERNAL_SERVER_ERROR,
                code: ErrorCode::InternalServerError,
                message: "Internal server error".into(),
            },
        }
    }
}

impl From<ToStrError> for ApiError {
    fn from(_: ToStrError) -> Self {
        ApiError {
            status: StatusCode::BAD_REQUEST,
            code: ErrorCode::InvalidHeader,
            message: "Invalid header value".into(),
        }
    }
}

impl From<chrono::ParseError> for ApiError {
    fn from(_: chrono::ParseError) -> Self {
        ApiError {
            status: StatusCode::BAD_REQUEST,
            code: ErrorCode::InvalidDate,
            message: "Invalid date header value".into(),
        }
    }
}

impl From<ParseIntError> for ApiError {
    fn from(_: ParseIntError) -> Self {
        ApiError {
            status: StatusCode::BAD_REQUEST,
            code: ErrorCode::InvalidHeader,
            message: "Invalid content length value".into(),
        }
    }
}
